/***********************************************************************
 * Breakpoints
 * 
 * compile with Kolmogorovs Max Flow implementation (fast): c++ -std=c++11 `Magick++-config --cxxflags --cppflags` -I./ -O2 -o breakpoints breakpoints.cpp `Magick++-config --ldflags --libs`  -lexternal-library-maxflow-shared
 * compile with Boost Max Flow implementation (slower, but header only): c++ -std=c++11 `Magick++-config --cxxflags --cppflags`  -O2 -o breakpoints breakpoints.cpp `Magick++-config --ldflags --libs`  
 * 
 * run: ./denoise [lambda] [sigma] [image filename]
 ************************************************************************/

//set to 1, if you want to compile with Kolmogorovs max flow implementation
#define USE_KOLMOGOROV 0

#include <Magick++.h> 
#include <iostream>
#include <cstdio>
#include <random>
#include <cmath>
#include <list>

#include <opengm/opengm.hxx>
#include <opengm/graphicalmodel/graphicalmodel.hxx>
#include <opengm/operations/adder.hxx>
#include <opengm/operations/minimizer.hxx>
#include <opengm/inference/graphcut.hxx>
#if USE_KOLMOGOROV
#include <opengm/inference/auxiliary/minstcutkolmogorov.hxx>
#else
#include <opengm/inference/auxiliary/minstcutboost.hxx>
#endif
//#include <opengm/inference/external/qpbo.hxx>

using namespace Magick;

//*****************************************
//** Typedefs
//*****************************************
// type used for values
typedef double ValueType;
// type used for indexing nodes and factors (default : size_t)
typedef size_t IndexType;
// type used for labels (default : size_t)
typedef size_t LabelType;
// operation used to combine terms
typedef opengm::Adder OpType;
// shortcut for explicite function
typedef opengm::ExplicitFunction<ValueType,IndexType,LabelType> ExplicitFunction;
// list of all function the model cal use (this trick avoids virtual methods) - here only one
typedef opengm::meta::TypeListGenerator<ExplicitFunction>::type FunctionTypeList;
// type used to define the feasible statespace
typedef opengm::DiscreteSpace<IndexType, LabelType> SpaceType;
// type of the model
typedef opengm::GraphicalModel<ValueType,OpType,FunctionTypeList,SpaceType> Model;
// type of the function identifier
typedef Model::FunctionIdentifier FunctionIdentifier;

#if USE_KOLMOGOROV
typedef opengm::external::MinSTCutKolmogorov<size_t, double> MinSTCutKolmogorov;
typedef opengm::GraphCut<Model, opengm::Minimizer, MinSTCutKolmogorov> MinCut;
#else
typedef opengm::MinSTCutBoost<size_t, double, opengm::PUSH_RELABEL> MinSTCutType;
typedef opengm::GraphCut<Model, opengm::Minimizer, MinSTCutType> MinCut;
#endif

//*****************************************
//** Declare functions 
//*****************************************
void constructModel(Image& _image, double _lambda, Model** _model, FunctionIdentifier& _lengthTermId);
//construct a model, that has only the data term
void constructModel0(Image& _image, Model** _model);
//construct a model, that has only the length term
void constructModel1(Image& _image, double _lambda, Model** _model, FunctionIdentifier& _lengthTermId);
void inferModel(Model* _model, double &_energy, int _W, int _H, std::vector<LabelType>* _labeling = 0, Image* _segmentation = 0);
void setLambda(Model* _model, double _lambda, FunctionIdentifier _lengthTermId);
std::vector<LabelType>::size_type diff(const std::vector<LabelType>& _labeling0, const std::vector<LabelType>& _labeling1);
void showLabeling(const std::vector<LabelType>& _labeling, int _W, int _H);

struct Interval
{
    double start;
    double end;
    double energ0;
    double energ1;
    std::vector<LabelType> x;
};

typedef std::list<Interval> IntervalList;

//*****************************************
//** main
//*****************************************
int main(int argc, char** argv) {
    InitializeMagick(*argv);
    
    //*****************************************
    //** Parameters
    //*****************************************
    double paramLambda = 1.0;
    double paramEpsilon = 0.1;
    int paramEpsilon2 = 1;
    std::string imageFilename = "garden.png";
    
    if (argc>1)
    {
        imageFilename = std::string(argv[1]);
    }
    if (argc>2)
    {
        paramEpsilon = std::stod(std::string(argv[2]));
    }
    if (argc>3)
    {
        paramEpsilon2 = std::stoi(std::string(argv[3]));
    }
    if (argc>4)
    {
        paramLambda = std::stod(std::string(argv[4]));
    }
    
    std::cout << "Using image in file " << imageFilename << std::endl;
    std::cout << "Parameter lambda = " << paramLambda << std::endl;
    std::cout << "Parameter epsilon = " << paramEpsilon << std::endl;
    std::cout << "Parameter epsilon2 = " << paramEpsilon2 << std::endl;
    
    try{
        //*****************************************
        //** Load Image
        //*****************************************
        
        Image image;
        image.read(imageFilename);

        int W = image.size().width();
        int H = image.size().height();

        // Show image
        Image tempImage = image;
        //tempImage.display();

        //*****************************************
        //** Build models
        //*****************************************
        Model* gm = 0;
        Model* gm0 = 0;
        Model* gm1 = 0;
        FunctionIdentifier lengthTermId = 0;
        FunctionIdentifier lengthTerm1Id = 0;
        std::cout << "Start building the model ... "<<std::endl;
        constructModel(image, paramLambda, &gm, lengthTermId);
        constructModel0(image, &gm0);
        constructModel1(image, 1, &gm1, lengthTerm1Id);
        std::cout << "building the model done. "<<std::endl;

        double energy = -1;
        Image segmentation = image;
        std::cout << "Start inferring ... "<<std::endl;
        inferModel(gm, energy, W,H,0,&segmentation);
        std::cout << "inferring done. "<<std::endl;

        // Show segmentation
        tempImage = segmentation;
        tempImage.display();

        //setting lambda to a different value without rebuilding the model.
        setLambda(gm, paramLambda*2, lengthTermId);
        std::cout << "Start inferring ... "<<std::endl;
        inferModel(gm, energy, W,H,0,&segmentation);
        std::cout << "inferring done. "<<std::endl;
        // Show segmentation (this segmentation should now have fewer connected components)
        tempImage = segmentation;
        tempImage.display();
        
        //@TODO insert your code for finding the breakpoints here.
    }
    catch( Exception &error_ ) 
    { 
        std::cout << "Caught exception: " << error_.what() << std::endl; 
        return 1; 
    } 


}


/***********************************************************************
 * Implementation other functions
 * 
 ************************************************************************/

void constructModel(Image& _image, double _lambda, Model** _model, FunctionIdentifier& _lengthTermId)
{
    int W = _image.size().width();
    int H = _image.size().height();
    
    //*****************************************
    //** Construct Model
    //*****************************************
    // Build empty Model
    LabelType numLabel = 2;
    std::vector<LabelType> numbersOfLabels(W*H,numLabel);
    Model* gm =  new Model(SpaceType(numbersOfLabels.begin(), numbersOfLabels.end()));
    // Add 1st order functions and factors to the model
    for (int y=0; y<H; ++y)
    {
        for (int x=0; x<W; ++x)
        {
            //for the data term, we are using simple dark/light foreground/background model
            ColorGray c = _image.pixelColor(x,y);
            IndexType linearIndex = x+W*y;
            // construct 1st order function
            const LabelType shape[] = {numLabel};
            ExplicitFunction f(shape, shape + 1);
            f(0) = c.shade();
            f(1) = 1-c.shade();
            // add function
            FunctionIdentifier id = gm->addFunction(f);
            // add factor
            IndexType variableIndex[] = {linearIndex};
            gm->addFactor(id, variableIndex, variableIndex + 1);
        }
    }
    const LabelType pairwiseParameter[] = {numLabel,numLabel};
    //add a explicit function with 2 variables as the horizontal regularizer
    ExplicitFunction c(pairwiseParameter, pairwiseParameter + 2);
    c(0,0) = 0;
    c(1,0) = _lambda;
    c(0,1) = _lambda;
    c(1,1) = 0;
    _lengthTermId = gm->addFunction(c);

    // add 2nd order functions for all variables neighbored on the grid
    for (int y=0; y<H; ++y)
    {
        for (int x=0; x<W; ++x)
        {
            IndexType linearIndex00 = x+W*y;
            IndexType linearIndex01 = x+W*(y+1);
            IndexType linearIndex10 = x+1+W*y;
            IndexType vars[]  = {linearIndex00,1}; 

            if(x+1<W){ //check for horizontal neighbor
                vars[1] =  linearIndex10;
                gm->addFactor(_lengthTermId, vars, vars + 2);
            } 
            if(y+1<H){ //check for vertical neighbor
                vars[1] =  linearIndex01;
                gm->addFactor(_lengthTermId, vars, vars + 2);
            } 
        }
    }
    *_model = gm;
}

void constructModel0(Image& _image, Model** _model)
{
    int W = _image.size().width();
    int H = _image.size().height();
    
    //*****************************************
    //** Construct Model
    //*****************************************
    // Build empty Model
    LabelType numLabel = 2;
    std::vector<LabelType> numbersOfLabels(W*H,numLabel);
    Model* gm =  new Model(SpaceType(numbersOfLabels.begin(), numbersOfLabels.end()));
    // Add 1st order functions and factors to the model
    for (int y=0; y<H; ++y)
    {
        for (int x=0; x<W; ++x)
        {
            //for the data term, we are using simple dark/light foreground/background model
            ColorGray c = _image.pixelColor(x,y);
            IndexType linearIndex = x+W*y;
            // construct 1st order function
            const LabelType shape[] = {numLabel};
            ExplicitFunction f(shape, shape + 1);
            f(0) = c.shade();
            f(1) = 1-c.shade();
            // add function
            FunctionIdentifier id = gm->addFunction(f);
            // add factor
            IndexType variableIndex[] = {linearIndex};
            gm->addFactor(id, variableIndex, variableIndex + 1);
        }
    }
    *_model = gm;
}

void constructModel1(Image& _image, double _lambda, Model** _model, FunctionIdentifier& _lengthTermId)
{
    int W = _image.size().width();
    int H = _image.size().height();
    
    //*****************************************
    //** Construct Model
    //*****************************************
    // Build empty Model
    LabelType numLabel = 2;
    std::vector<LabelType> numbersOfLabels(W*H,numLabel);
    Model* gm =  new Model(SpaceType(numbersOfLabels.begin(), numbersOfLabels.end()));
    const LabelType pairwiseParameter[] = {numLabel,numLabel};
    //add a explicit function with 2 variables as the horizontal regularizer
    ExplicitFunction c(pairwiseParameter, pairwiseParameter + 2);
    c(0,0) = 0;
    c(1,0) = _lambda;
    c(0,1) = _lambda;
    c(1,1) = 0;
    _lengthTermId = gm->addFunction(c);

    // add 2nd order functions for all variables neighbored on the grid
    for (int y=0; y<H; ++y)
    {
        for (int x=0; x<W; ++x)
        {
            IndexType linearIndex00 = x+W*y;
            IndexType linearIndex01 = x+W*(y+1);
            IndexType linearIndex10 = x+1+W*y;
            IndexType vars[]  = {linearIndex00,1}; 

            if(x+1<W){ //check for horizontal neighbor
                vars[1] =  linearIndex10;
                gm->addFactor(_lengthTermId, vars, vars + 2);
            } 
            if(y+1<H){ //check for vertical neighbor
                vars[1] =  linearIndex01;
                gm->addFactor(_lengthTermId, vars, vars + 2);
            } 
        }
    }
    *_model = gm;
}

void inferModel(Model* _model, double &_energy, int _W, int _H, std::vector<LabelType>* _labeling, Image* _segmentation)
{
    // Infer with Graph Cut
    MinCut::Parameter para(1000000); //scale capacities, only needed if we use integer capacities
    MinCut minCut(*_model,para);
    minCut.infer();
    _energy = minCut.value();
    std::vector<LabelType>* l_MinCut = 0;
    if (_labeling!=0)
    {
        l_MinCut = _labeling;
    }
    else
    {
        l_MinCut = new std::vector<LabelType>();
    }
    minCut.arg(*l_MinCut);
    if (_segmentation!=0)
    {
        assert(_segmentation->size().width()==_W);
        assert(_segmentation->size().height()==_H);
        for (int y=0; y<_H; ++y)
        {
            for (int x=0; x<_W; ++x)
            {
                int linearIndex = x+_W*y;
                ColorMono c((*l_MinCut)[linearIndex]!=0);
                _segmentation->pixelColor(x,y,c);
            }
        }
    }
    if (_labeling==0)
    {
        delete l_MinCut;
    }
}

void setLambda(Model* _model, double _lambda, FunctionIdentifier _lengthTermId)
{
    ExplicitFunction& c = _model->getFunction<ExplicitFunction>(_lengthTermId);
    c(1,0) = _lambda;
    c(0,1) = _lambda;
}
   
std::vector<LabelType>::size_type diff(const std::vector<LabelType>& _labeling0, const std::vector<LabelType>& _labeling1)
{
    assert(_labeling0.size()==_labeling1.size());
    std::vector<LabelType>::size_type result = 0;
    for (std::vector<LabelType>::size_type i=0; i<_labeling0.size(); ++i)
    {
	if (_labeling0[i]!=_labeling1[i])
	    ++result;
    }
    return result;
}

void showLabeling(const std::vector<LabelType>& _labeling, int _W, int _H)
{
    Image image(Geometry(_W, _H), ColorMono(false));
    for (int y=0; y<_H; ++y)
    {
	for (int x=0; x<_W; ++x)
	{
	    int linearIndex = x+_W*y;
	    ColorMono c(_labeling[linearIndex]!=0);
	    image.pixelColor(x,y,c);
	}
    }
    image.display();
}
